# pylint: disable=no-value-for-parameter
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments

import logging

import click
import uvicorn

from ant31box.config import LOG_LEVELS, Config
from ant31box.config import config as confload
from ant31box.init import init

LEVEL_CHOICES = click.Choice(list(LOG_LEVELS.keys()))
logger = logging.getLogger("ant31box.info")


def run_server(config: Config):
    logger.info("Starting server")
    click.echo(f"{config.server.model_dump()}")
    init(config.conf, "fastapi")
    uvicorn.run(
        config.server.server,
        host=config.server.host,
        port=config.server.port,
        log_level=config.logging.level,
        # log_config=config.logging.log_config,
        use_colors=config.logging.use_colors,
        reload=config.server.reload,
        factory=True,
    )


# pylint: disable=no-value-for-parameter
# pylint: disable=too-many-arguments
@click.command(context_settings={"auto_envvar_prefix": "FASTAPI"})
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    default=None,
    help="Configuration file in YAML format.",
    show_default=True,
)
@click.option(
    "--host",
    type=str,
    default=None,
    help="Address of the server",
    show_default=True,
)
@click.option(
    "--log-config",
    type=click.Path(exists=True),
    default=None,
    help="Logging configuration file. Supported formats: .ini, .json, .yaml.",
    show_default=True,
)
@click.option(
    "--log-level",
    type=LEVEL_CHOICES,
    default="info",
    help="Log level.",
    show_default=True,
)
@click.option(
    "--use-colors/--no-use-colors",
    is_flag=True,
    default=True,
    help="Enable/Disable colorized logging.",
)
@click.option(
    "--port",
    default=None,
    type=int,
    help="Port to listen on",
)
def server(
    config: str,
    host: str,
    port: int,
    use_colors: bool,
    log_level: str,
    log_config: str,
) -> None:
    _config = confload(config)
    if host:
        _config.server.host = host
    if port:
        _config.server.port = port
    if log_level:
        _config.logging.level = log_level
    if log_config:
        _config.logging.log_config = log_config
    if use_colors is not None:
        _config.logging.use_colors = use_colors
    if host:
        _config.conf.server.host = host

    run_server(_config)



# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument
# pylint: disable=too-few-public-methods
import logging
from typing import Any

import temporalio.client
from fastapi import APIRouter
from temporalloop.importer import import_from_string

from ant31box.models import AsyncResponse
from ant31box.server.exception import ResourceNotFound
from ant31box.temporal.client import tclient

router = APIRouter(prefix="/api/v1/status", tags=["antbed", "status"])

logger = logging.getLogger(__name__)


async def get_handler_from_ar(
    ar: AsyncResponse,
) -> tuple[temporalio.client.WorkflowHandle[Any, Any], Any]:
    """
    Retrieve the workflow handler and workflow object from the AsyncResponse.

    :param ar: The AsyncResponse object containing the job details.
    :return: A tuple containing the workflow handler and workflow object.
    """
    return await get_handler(ar.payload.jobs[0].uuid, ar.payload.jobs[0].uuid)


async def get_handler(
    workflow_id: str,
    workflow_name: str,
) -> tuple[temporalio.client.WorkflowHandle[Any, Any], Any]:
    """
    Retrieve the workflow handler and workflow object for the given workflow ID and name.

    :param workflow_id: The ID of the workflow.
    :param workflow_name: The name of the workflow.
    :return: A tuple containing the workflow handler and workflow object.
    """
    workflow = import_from_string(workflow_name)
    # Retrieve running workflow handler
    client = await tclient()
    return (
        client.get_workflow_handle_for(workflow_id=workflow_id, workflow=workflow.run),
        workflow,
    )


@router.post("/", response_model=AsyncResponse)
async def status(ar: AsyncResponse) -> AsyncResponse:
    """
    Retrieve the status of the workflow and update the AsyncResponse object.

    :param ar: The AsyncResponse object containing the job details.
    :return: The updated AsyncResponse object with the workflow status and result.
    """
    workflow_id = ar.payload.jobs[0].uuid
    handler, _ = await get_handler_from_ar(ar)
    describe = await handler.describe()
    j = ar.payload.jobs[0]
    if not describe.status:
        raise ResourceNotFound("Workflow not found", {"workflow_id": workflow_id})
    j.status = describe.status.name
    if describe.status == temporalio.client.WorkflowExecutionStatus.COMPLETED:
        j.result = await handler.result()
    ar.gen_signature()
    return ar



from typing import ClassVar

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette_exporter import handle_metrics
from starlette_exporter.middleware import PrometheusMiddleware
from uvicorn.importer import import_from_string
from uvicorn.middleware.proxy_headers import ProxyHeadersMiddleware

from ant31box.config import Config, FastAPIConfigSchema, config
from ant31box.init import init

from .middlewares.errors import catch_exceptions_middleware
from .middlewares.process_time import add_process_time_header
from .middlewares.token import TokenAuthMiddleware


def cors(server: "Server"):
    server.app.add_middleware(
        CORSMiddleware,
        allow_origin_regex=server.config.cors.allow_origin_regex,
        allow_origins=server.config.cors.allow_origins,
        allow_credentials=server.config.cors.allow_credentials,
        allow_methods=server.config.cors.allow_methods,
        allow_headers=server.config.cors.allow_headers,
    )


def add_process_time_header_m(server: "Server"):
    server.app.middleware("http")(add_process_time_header)


def token_auth(server: "Server"):
    if server.config.token is not None:
        server.config.token_auth.token = server.config.token

    server.app.add_middleware(
        TokenAuthMiddleware,
        token=server.config.token_auth.token,
        parameter=server.config.token_auth.parameter,
        header_name=server.config.token_auth.header_name,
        skip_paths=server.config.token_auth.skip_paths,
    )


def catch_exceptions(server: "Server"):
    server.app.middleware("http")(catch_exceptions_middleware)


def prometheus(server: "Server"):
    server.app.add_middleware(PrometheusMiddleware, app_name=server.appname)
    server.app.add_route("/metrics", handle_metrics)


def proxy_headers(server: "Server"):
    server.app.add_middleware(ProxyHeadersMiddleware)


AVAILABLE_MIDDLEWARES = {
    "cors": "ant31box.server.server:cors",
    "tokenAuth": "ant31box.server.server:token_auth",
    "catchExceptions": "ant31box.server.server:catch_exceptions",
    "prometheus": "ant31box.server.server:prometheus",
    "proxyHeaders": "ant31box.server.server:proxy_headers",
    "addProcessTimeHeader": "ant31box.server.server:add_process_time_header_m",
}
DEFAULT_MIDDLEWARES = {"catchExceptions", "prometheus", "proxyHeaders", "addProcessTimeHeader"}
AVAILABLE_ROUTERS = {"ant31box.server.api.info:router", "ant31box.server.api.debug:router"}


class Server:
    _available_middlewares: ClassVar[dict[str, str]] = AVAILABLE_MIDDLEWARES
    _default_middlewares: ClassVar[set[str]] = DEFAULT_MIDDLEWARES
    _default_routers: ClassVar[set[str]] = AVAILABLE_ROUTERS

    _routers: ClassVar[set[str]] = set()
    _middlewares: ClassVar[set[str]] = set()

    def __init__(self, conf: FastAPIConfigSchema, appname="ant31box", appenv="dev"):
        self.config: FastAPIConfigSchema = conf
        self.app: FastAPI = FastAPI()
        self.appname = appname
        self.appenv = appenv

        defaultm = self._default_middlewares
        if self.config.middlewares_replace_default is not None:
            defaultm = self.config.middlewares_replace_default
        middlewares = set(self.config.middlewares).union(defaultm).union(self._middlewares)
        self.load_middlewares(middlewares)

        defaultr = self._default_routers
        if self.config.routers_replace_default is not None:
            defaultr = self.config.routers_replace_default
        routers = set(self.config.routers).union(defaultr).union(self._routers)
        self.load_routers(routers)

    @classmethod
    def set_available_middlewares(cls, middlewares: dict[str, str]):
        cls._available_middlewares = middlewares

    @classmethod
    def add_available_middlewares(cls, middlewares: dict[str, str]):
        cls._available_middlewares.update(middlewares)

    def load_middlewares(self, middlewares: set[str]):
        loaded = set()
        for middleware in middlewares:
            importname = middleware
            if middleware in self._available_middlewares:
                importname = self._available_middlewares[middleware]
            if importname in loaded:
                # already loaded
                continue
            callback = import_from_string(importname)
            callback(self)
            loaded.add(importname)

    def load_routers(self, routers: set[str]):
        debugr = "ant31box.server.api.debug:router"
        if self.appenv in ["prod", "production"] and debugr in routers:
            routers.remove(debugr)
        for router in routers:
            self.app.include_router(import_from_string(router))


def serve_from_config(conf: Config, server_class: type[Server] = Server) -> FastAPI:
    init(conf.conf, "fastapi")
    if not issubclass(server_class, Server):
        raise TypeError(f"server must be a subclass or instance of {Server}")

    server = server_class(conf.server, conf.name, conf.app.env)
    return server.app


# override this method to use a different server class/config
def serve() -> FastAPI:
    return serve_from_config(config(), Server)



import hashlib
import logging
from email.message import EmailMessage
from io import IOBase
from pathlib import Path
from typing import Literal
from urllib.parse import unquote, urlparse

import aiofiles
import aioshutil as shutil
from pydantic import BaseModel, ConfigDict, Field

from ant31box.client.base import BaseClient
from ant31box.config import S3ConfigSchema
from ant31box.s3 import S3URL, S3Client

# create a temporary directory using the context manager


logger: logging.Logger = logging.getLogger(__name__)


class FileInfo(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    filename: str = Field(default="")
    content: IOBase | None = Field(default=None)
    path: Path | str | None = Field(default=None)
    source: str | Path | None = Field(default=None)
    metadata: dict[str, str] | None = Field(default=None)


class DownloadClient(BaseClient):
    def __init__(self, s3_config: S3ConfigSchema | None = None) -> None:
        super().__init__(endpoint="", client_name="filedl")
        self.s3 = None
        if s3_config is not None:
            self.set_s3(s3_config)

    def set_s3(self, s3_config: S3ConfigSchema) -> None:
        self.s3 = S3Client(s3_config)

    def _gen_sha(self, content: bytes, source_path: str, dest_dir: str, filename: str) -> str:
        path = Path(source_path)
        hashsha = hashlib.sha256(content)
        suffix = path.suffix
        filename = hashsha.hexdigest() + suffix
        return str(Path(dest_dir).joinpath(filename))

    async def copy_local_file(
        self, source_path: str, dest_dir: str | Path = "", output: str | Path | IOBase = ""
    ) -> FileInfo:
        filename = Path(source_path).name
        # Write output
        # if output is a IOBase, write the content to it and return it
        if output and isinstance(output, IOBase):
            # Read input
            async with aiofiles.open(source_path, "rb") as fopen:
                output.write(await fopen.read())
                return FileInfo(content=output, filename=filename, source=source_path)

        # if output is a string, write the content to that file and return
        elif output and isinstance(output, Path | str):
            dest_path = output
        else:
            dest_path = Path(dest_dir).joinpath(filename)
        await shutil.copyfile(source_path, dest_path)
        return FileInfo(filename=filename, path=dest_path, source=source_path)

    def headers(
        self, content_type: Literal["json", "form"] | str | None = None, extra: dict[str, str] | None = None
    ) -> dict[str, str]:
        headers = {
            "Accept": "*/*",
        }
        if extra is not None:
            headers.update(extra)
        return super().headers(content_type=content_type, extra=headers)

    async def download_file(
        self, url: str, source_path: str, dest_dir: str | Path = "", output: str | Path | IOBase = ""
    ) -> FileInfo:
        resp = await self.session.get(url, headers=self.headers())
        resp.raise_for_status()

        filename: str = ""

        content_disposition = resp.headers.get("Content-Disposition")
        if content_disposition:
            msg = EmailMessage()
            msg["Content-Disposition"] = content_disposition
            _, params = msg.get_content_type(), msg["Content-Disposition"].params
            filename = params.get("filename", "")
        if not filename:
            filename = unquote(Path(source_path).name)

        content = await resp.content.read()

        if output and isinstance(output, IOBase):
            output.write(content)
            fd = FileInfo(content=output, filename=filename, source=url)
            return fd
        if output and isinstance(output, Path | str):
            dest_path = output
        else:
            dest_path = Path(dest_dir).joinpath(filename)
        async with aiofiles.open(dest_path, "wb") as fopen:
            await fopen.write(content)
        fd = FileInfo(filename=filename, path=dest_path, source=url)
        return fd

    def download_s3(self, source: str, dest_dir: str | Path = "", output: str | Path | IOBase = "") -> FileInfo:
        s3url = S3URL(url=source)
        fd = FileInfo(source=source, filename=s3url.filename, metadata=s3url.to_dict())
        if self.s3 is None:
            raise AttributeError("S3 client is not set")
        if not output:
            output = Path(dest_dir).joinpath(s3url.filename)
        if output and hasattr(output, "write"):
            # write to file-like object
            self.s3.download_file(s3url=s3url.to_model(), dest=output)
            fd.content = output
        if output and isinstance(output, Path | str):
            # write to file on disk
            with open(str(output), "wb") as fileobj:
                self.s3.download_file(s3url=s3url.to_model(), dest=fileobj)
            fd.path = output
        return fd

    async def download(self, source: str, dest_dir: str | Path = "", output: str | Path | IOBase = "") -> FileInfo:
        """
        Determine the protocol to fetch the document:
        file://,
        http://,
        s3:// ...
        """
        parsedurl = urlparse(source)
        logger.info("download %s, %s", source, parsedurl.path)
        if parsedurl.scheme in ["file", ""]:
            return await self.copy_local_file(source_path=parsedurl.path, dest_dir=dest_dir, output=output)
        if parsedurl.scheme in ["http", "https"]:
            return await self.download_file(url=source, source_path=parsedurl.path, dest_dir=dest_dir, output=output)
        if parsedurl.scheme in ["s3"]:
            return self.download_s3(source=source, dest_dir=dest_dir, output=output)
        raise AttributeError(f"Unsupported file source: scheme={parsedurl.scheme} - path={parsedurl.path}")



from functools import cache

from ant31box.client.filedl import DownloadClient


@cache
def filedl_client(key: str = "") -> DownloadClient:
    """
    Create a EPostClient instance with the given key
    It cache the answer for the same key
    use filedl_client.cache_clear() to clear the cache
    """
    _ = key
    return DownloadClient()



from collections.abc import Sequence

from starlette.datastructures import URL, Headers, QueryParams
from starlette.responses import JSONResponse
from starlette.types import ASGIApp, Receive, Scope, Send

from ..exception import UnauthorizedAccess


class TokenAuthMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        token: str = "",
        parameter: str | None = "auth_token",
        header_name: str | None = "token",
        skip_paths: Sequence[str] = [
            "/api/v1/auth/login",
            "/api/v1/auth/register",
            "/docs",
            "/openapi.json",
            "/redoc",
            "/metrics",
            "/health",
        ],
    ) -> None:
        self.token: str = token
        self.app = app
        self.parameter: str | None = parameter
        self.header_name: str | None = header_name
        self.skip_paths: set[str] = set(skip_paths)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        headers = Headers(scope=scope)
        url = URL(scope=scope)
        path = url.path
        params = QueryParams(scope=scope)

        # if token is not set or path is in skip_paths then skip
        # if token is set and matches with parameter or header then allow
        if (
            (not self.token or path in self.skip_paths)
            or (self.parameter and self.parameter in params and params[self.parameter] == self.token)
            or (self.header_name and self.header_name in headers and headers[self.header_name] == self.token)
        ):
            await self.app(scope, receive, send)
            return

        # if token is set and doesn't match then return UnauthorizedAccess error
        error = UnauthorizedAccess("NoAuth")
        await JSONResponse({"error": error.to_dict()}, status_code=error.status_code)(scope, receive, send)
        return



#!/usr/bin/env python3
import importlib
from collections.abc import Mapping
from typing import Any


# from uvicorn.importer (BSD 3-Clause https://github.com/encode/uvicorn/blob/master/LICENSE.md)
class ImportFromStringError(Exception):
    pass


# From uvicorn.importer (BSD 3-Clause https://github.com/encode/uvicorn/blob/master/LICENSE.md)
def import_from_string(import_str: Any) -> Any:
    if not isinstance(import_str, str):
        return import_str

    module_str, _, attrs_str = import_str.partition(":")
    if not module_str or not attrs_str:
        message = 'Import string "{import_str}" must be in format "<module>:<attribute>".'
        raise ImportFromStringError(message.format(import_str=import_str))

    try:
        module = importlib.import_module(module_str)
    except ModuleNotFoundError as exc:
        if exc.name != module_str:
            raise exc from None
        message = 'Could not import module "{module_str}".'
        raise ImportFromStringError(message.format(module_str=module_str))

    instance = module
    try:
        for attr_str in attrs_str.split("."):
            instance = getattr(instance, attr_str)
    except AttributeError:
        message = 'Attribute "{attrs_str}" not found in module "{module_str}".'
        raise ImportFromStringError(message.format(attrs_str=attrs_str, module_str=module_str))

    return instance


def deepmerge(dct, merge_dct):
    """Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k, _ in merge_dct.items():
        if k in dct and isinstance(dct[k], Mapping) and isinstance(merge_dct[k], Mapping):
            deepmerge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]



#!/usr/bin/env python3

import pathlib

import sentry_sdk
from sentry_sdk.integrations.fastapi import FastApiIntegration
from sentry_sdk.integrations.starlette import StarletteIntegration

from ant31box.config import ConfigSchema, SentryConfigSchema


def init_sentry(config: SentryConfigSchema, integration_app: str = ""):
    if config.dsn:
        integrations = []
        if integration_app == "fastapi":
            integrations = [
                StarletteIntegration(),
                FastApiIntegration(),
            ]
        sentry_sdk.init(  # pylint: disable=abstract-class-instantiated
            dsn=config.dsn,
            integrations=integrations,
            traces_sample_rate=config.traces_sample_rate,
            environment=config.environment,
        )


def _create_tmp_dir(promdir: str) -> None:
    pathlib.Path(promdir).mkdir(parents=True, exist_ok=True)


def init(config: ConfigSchema, app: str = "fastapi"):
    _create_tmp_dir(config.app.prometheus_dir)
    init_sentry(config.sentry, app)



#!/usr/bin/env python3
import json
import logging
from typing import Any, Literal, TypeVar
from urllib.parse import ParseResult, urlparse

import aiohttp
from pydantic import BaseModel, ConfigDict, Field

from ant31box.version import VERSION

logger = logging.getLogger(__name__)

T = TypeVar("T")  # , bound="ClientConfig")
# pylint: disable=protected-access


class ClientConfig(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")
    endpoint: str = Field(default="http://localhost:8080")
    client_name: str = Field(default="client")
    verify_tls: bool = Field(default=True)
    session_args: tuple[list, dict[str, Any]] = Field(default=([], {}))


class BaseClient:
    def __init__(
        self,
        endpoint: str,
        verify_tls: bool = True,
        session_args: tuple[list, dict[str, Any]] = ([], {}),
        client_name: str = "client",
    ) -> None:
        self._session: aiohttp.ClientSession | None = None
        self.client_config = ClientConfig(
            endpoint=endpoint, verify_tls=verify_tls, session_args=session_args, client_name=client_name
        )
        self._endpoint: ParseResult = self._configure_endpoint(self.client_config.endpoint)
        self._headers: dict[str, str] = {
            "Content-Type": "application/json",
            "User-Agent": f"ant31box-cli/{self.client_config.client_name}-{VERSION.app_version}",
        }

    @property
    def endpoint(self) -> ParseResult:
        return self._endpoint

    @endpoint.setter
    def endpoint(self, value: str) -> None:
        self._endpoint = self._configure_endpoint(value)

    @property
    def ssl_mode(self) -> bool:
        return self.client_config.verify_tls

    def close(self):
        """
        Close aiohttp.ClientSession.

        This is useful to be called manually in tests if each test when each test uses a new loop. After close, new
        requests will automatically create a new session.

        Note: We need a sync version for `__del__` and `aiohttp.ClientSession.close()` is async even though it doesn't
        have to be.
        """
        if self._session:
            if not self._session.closed:
                # Older aiohttp does not have _connector_owner
                if not hasattr(self._session, "_connector_owner") or self._session._connector_owner:
                    try:
                        if self._session._connector:
                            self._session._connector._close()  # New version returns a coroutine in close() as warning
                    except Exception:  # pylint: disable=broad-exception-caught
                        if self._session._connector:
                            self._session._connector.close()
                self._session._connector = None
            self._session = None

    @property
    def session(self) -> aiohttp.ClientSession:
        """An instance of aiohttp.ClientSession"""
        if not self._session or self._session.closed or not self._session._loop or self._session._loop.is_closed():
            self._session = aiohttp.ClientSession(
                *self.client_config.session_args[0], **self.client_config.session_args[1]
            )
        return self._session

    # pylint: disable=too-many-arguments
    async def log_request(self, resp: aiohttp.ClientResponse) -> None:
        try:
            raw = await resp.read()
        except aiohttp.ClientPayloadError:
            raw = b""
        try:
            logger.debug(
                json.dumps(
                    {
                        "query": {
                            "url": str(resp.request_info.url),
                            "method": resp.request_info.method,
                            "headers": dict(resp.request_info.headers.items()),
                        },
                        "response": {
                            "headers": dict(resp.headers.items()),
                            "status": resp.status,
                            "raw": raw.decode("utf-8"),
                        },
                    },
                    default=str,
                )
            )
        except Exception:  # pylint: disable=broad-exception-caught
            pass

    def _url(self, path: str, endpoint: str = "") -> str:
        """Construct the url from a relative path"""
        if endpoint:
            ep = urlparse(endpoint)
        else:
            ep = self.endpoint
        return ep.geturl() + path

    def _configure_endpoint(self, endpoint: str) -> ParseResult:
        return urlparse(endpoint)

    def headers(
        self,
        content_type: Literal["json", "form"] | str | None = None,
        extra: dict[str, str] | None = None,
    ) -> dict[str, str]:
        headers: dict[str, str] = {}
        headers.update(self._headers)

        if content_type == "json":
            headers["Content-Type"] = "application/json"
        elif content_type == "form":
            headers["Content-Type"] = "application/x-www-form-urlencoded"
        elif content_type:
            headers["Content-Type"] = content_type

        if extra:
            headers.update(extra)

        return headers



import logging
from io import IOBase
from pathlib import Path
from typing import BinaryIO

import boto3
from botocore.client import Config

from ant31box.config import S3ConfigSchema
from ant31box.models import S3URL, S3Dest

logger: logging.Logger = logging.getLogger(__name__)


class S3Client:
    def __init__(self, options: S3ConfigSchema, bucket: str = "", prefix: str = ""):
        kwargs: dict = self._boto_args(options)
        self.options = options
        self.client = boto3.resource("s3", **kwargs)
        if not bucket:
            bucket = options.bucket
        if not prefix:
            prefix = options.prefix
        self.bucket: str = bucket
        self.prefix: str = prefix

    @staticmethod
    def _boto_args(options: S3ConfigSchema):
        kwargs: dict = {}
        if options.endpoint:
            kwargs["endpoint_url"] = options.endpoint
        if options.region:
            kwargs["region_name"] = options.region
        if options.access_key:
            kwargs["aws_access_key_id"] = options.access_key
        if options.secret_key:
            kwargs["aws_secret_access_key"] = options.secret_key
        kwargs["config"] = Config(signature_version="s3v4")
        return kwargs

    def buildpath(self, filename: str, dest: str = ""):
        if not dest:
            dest = Path(filename).name
        return f"{self.prefix}{dest}"

    def upload_file(self, filepath: str | IOBase | BinaryIO, dest: str = "") -> S3Dest:
        if isinstance(filepath, str):
            path = self.buildpath(filepath, dest)
            logger.info("upload s3 bucket='%s' file='%s' dest='%s'", self.bucket, filepath, path)
            self.client.Bucket(self.bucket).upload_file(filepath, path)
        else:
            path = dest
            logger.info("upload s3 bucket='%s' file='%s' dest='%s'", self.bucket, filepath, path)
            self.client.Bucket(self.bucket).upload_fileobj(filepath, path)

        return S3URL(bucket=self.bucket, key=path, region=self.options.region).to_model()

    def s3url(self, path: str, strip: bool = True) -> S3URL:
        if strip:
            path = path.lstrip("/")
        return S3URL(bucket=self.bucket, key=path)

    def download_file(self, s3url: S3Dest, dest: str | Path | IOBase | BinaryIO) -> str | IOBase | BinaryIO:
        logger.info("download uri='%s', dest='%s'", s3url.url, dest)
        if isinstance(dest, str | Path):
            self.client.Bucket(s3url.bucket).download_file(s3url.key, str(dest))
        else:
            self.client.Bucket(s3url.bucket).download_fileobj(s3url.key, dest)
        return dest

    def copy_s3_to_s3(
        self,
        *,
        src_bucket: str,
        src_path: str,
        dest_bucket: str,
        dest_prefix: str = "",
        name_only: bool = False,
    ) -> tuple[S3Dest, S3Dest]:
        copy_source = {
            "Bucket": src_bucket,
            "Key": src_path,
        }

        if not name_only:
            dest_path = f"{dest_prefix}{src_path}"
        else:
            dest_path = f"{dest_prefix}{Path(src_path).name}"

        self.client.meta.client.copy(copy_source, dest_bucket, dest_path)

        return (
            S3URL(bucket=src_bucket, key=src_path, region=self.options.region).to_model(),
            S3URL(bucket=dest_bucket, key=dest_path, region=self.options.region).to_model(),
        )



#!/usr/bin/env python3
import asyncio
import functools


def make_sync(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return wrapper



from fastapi import Form


def form_body(prefix: str = ""):
    def inner(cls):
        parameters = []
        for arg in cls.__signature__.parameters.values():
            parameters.append(arg.replace(default=Form(arg.default, alias=f"{prefix}{arg.name}")))
        cls.__signature__ = cls.__signature__.replace(parameters=parameters)
        return cls

    return inner



import time
from collections.abc import Awaitable, Callable

from fastapi import Request, Response


async def add_process_time_header(request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
    start_time = time.time()
    response = await call_next(request)
    process_time = time.time() - start_time
    response.headers["X-Process-Time"] = str(process_time)
    return response



import click

from .default_config import default_config
from .server import server
from .version import version


@click.group()
@click.pass_context
def cli(ctx: click.Context) -> None:
    ctx.ensure_object(dict)


def main():
    # start the FastAPI server
    cli.add_command(server)
    # Display version
    cli.add_command(version)
    # Show default config
    cli.add_command(default_config)

    # Parse cmd-line arguments and options
    # pylint: disable=no-value-for-parameter
    cli()


if __name__ == "__main__":
    main()



import logging
import traceback
from collections.abc import Awaitable, Callable

from fastapi import Request, Response
from starlette.responses import JSONResponse

from ..exception import APIException

logger = logging.getLogger(__name__)


async def catch_exceptions_middleware(
    request: Request, call_next: Callable[[Request], Awaitable[Response]]
) -> Response:
    try:
        return await call_next(request)
    except APIException as error:
        # you probably want some kind of logging here
        logger.error(error)
        logger.error(traceback.format_exc())
        return JSONResponse({"error": error.to_dict()}, status_code=error.status_code)

    except Exception as err:  # pylint: disable=broad-except
        logger.error(err)
        logger.error(traceback.format_exc())
        error = APIException("Internal server error", {})
        return JSONResponse({"error": error.to_dict()}, status_code=error.status_code)



import hashlib
import hmac
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

from pydantic import BaseModel, Field


class S3Dest(BaseModel):
    bucket: str = Field(...)
    key: str = Field(...)
    url: str = Field(default="")
    region: str = Field(default="")

    def to_s3url(self) -> "S3URL":
        return S3URL(url=self.url, bucket=self.bucket, key=self.key, region=self.region)


class S3URL:
    def __init__(self, url: str = "", bucket: str = "", key: str = "", region: str = ""):
        self.parse_url(url)
        if bucket:
            self.bucket = bucket
        if key:
            self.key = key
        self.region = region

    def to_dict(self):
        return {
            "bucket": self.bucket,
            "key": self.key,
            "url": self.url,
            "region": self.region,
        }

    def to_model(self) -> S3Dest:
        return S3Dest(bucket=self.bucket, key=self.key, url=self.url, region=self.region)

    @property
    def url(self):
        return f"s3://{self.bucket}/{self.key}"

    def parse_url(self, value: str):
        if not value:
            return
        parsed = urlparse(value, allow_fragments=False)
        if parsed.scheme != "s3":
            raise ValueError(f"Invalid URL scheme {parsed.scheme}, must be 's3'")
        self.bucket = parsed.netloc
        self.key = parsed.path.lstrip("/")

    @property
    def filename(self) -> str:
        return Path(self.key).name


class Job(BaseModel):
    uuid: str = Field(...)
    name: str = Field(...)
    status: str | None = Field(default=None)
    result: dict[str, Any] = Field(default={})


class JobList(BaseModel):
    jobs: list[Job] = Field([])


class AsyncResponse(BaseModel):
    payload: JobList = Field(default=JobList(jobs=[]))
    signature: str | None = Field(default=None)
    _secret_key: bytes = b"NhqPtmfle3R5aNIP1e3uzjonClVN65XAbvqqM6A7H5fATj0j"

    def gen_signature(self):
        self.signature = hmac.new(self._secret_key, self.payload.model_dump_json().encode(), hashlib.sha256).hexdigest()
        return self.signature

    def check_signature(self):
        expect = hmac.new(self._secret_key, self.payload.model_dump_json().encode(), hashlib.sha256).hexdigest()
        return expect != self.signature



# pylint: disable=no-name-in-module
# pylint: disable=too-few-public-methods
import asyncio
import logging

from fastapi import APIRouter

from ant31box.server.exception import Forbidden

router = APIRouter()

logger = logging.getLogger(__name__)
router = APIRouter(prefix="/debug", tags=["debug"])


@router.get("/error", tags=["debug"])
async def gen_error():
    raise Forbidden("test")


@router.get("/error_uncatched", tags=["debug"])
async def gen_error_uncatch():
    raise Exception()


@router.get("/slow", tags=["debug"])
async def slow_req():
    await asyncio.sleep(5)
    return {"ok": 200}



import json

import click

from ant31box.version import VERSION, Version


@click.command()
@click.option("--output", "-o", default="json", type=click.Choice(["json", "text"]))
@click.pass_context
def version(ctx: click.Context, output: str, ver: Version = VERSION) -> None:
    if output == "json":
        click.echo(json.dumps(ver.to_dict(), indent=2))
    else:
        click.echo(ver.text())
    ctx.exit()



#!/usr/bin/env python3
# https://github.com/encode/uvicorn/blob/master/uvicorn/importer.py
# Copyright © 2017-present, [Encode OSS Ltd](https://www.encode.io/).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import importlib
from typing import Any


class ImportFromStringError(Exception):
    pass


def import_from_string(import_str: Any) -> Any:
    if not isinstance(import_str, str):
        return import_str

    module_str, _, attrs_str = import_str.partition(":")
    if not module_str or not attrs_str:
        message = 'Import string "{import_str}" must be in format "<module>:<attribute>".'
        raise ImportFromStringError(message.format(import_str=import_str))

    try:
        module = importlib.import_module(module_str)
    except ImportError as exc:
        if exc.name != module_str:
            raise exc from None
        message = 'Could not import module "{module_str}".'
        raise ImportFromStringError(message.format(module_str=module_str)) from exc

    instance = module
    try:
        for attr_str in attrs_str.split("."):
            instance = getattr(instance, attr_str)
    except AttributeError as exc:
        message = 'Attribute "{attrs_str}" not found in module "{module_str}".'
        raise ImportFromStringError(message.format(attrs_str=attrs_str, module_str=module_str)) from exc

    return instance



import click
import yaml

from ant31box.config import LOGGING_CONFIG, LoggingConfigSchema, config


@click.command()
def default_config() -> None:
    conf = config()
    conf.conf.logging = LoggingConfigSchema(log_config=LOGGING_CONFIG)
    click.echo(yaml.dump(conf.conf.model_dump(), default_flow_style=False))



from typing import Any

from fastapi import HTTPException


class APIException(HTTPException):
    status_code: int = 500
    errorcode: str = "internal-error"
    payload: dict[str, Any]

    def __init__(self, message: str, payload: dict[str, Any] | None = None) -> None:
        if not payload:
            payload = {}
        self.payload: dict[str, Any] = dict(payload)
        self.message = message
        super().__init__(status_code=self.status_code, detail=self.to_dict())

    def to_dict(self) -> dict[str, Any]:
        return {
            "code": self.errorcode,
            "message": self.message,
            "details": self.payload,
        }

    def __str__(self) -> str:
        return self.message


class InvalidUsage(APIException):
    status_code = 400
    errorcode = "invalid-usage"


class InvalidParams(APIException):
    status_code = 422
    errorcode = "invalid-parameters"


class ResourceNotFound(APIException):
    status_code = 404
    errorcode = "resource-not-found"


class Forbidden(APIException):
    status_code = 403
    errorcode = "forbidden"


class UnauthorizedAccess(APIException):
    status_code = 401
    errorcode = "unauthorized-access"


class Unsupported(APIException):
    status_code = 501
    errorcode = "unsupported"


class Unexpected(APIException):
    status_code = 500
    errorcode = "unexpected-error"



# pylint: disable=no-name-in-module
# pylint: disable=too-few-public-methods
import logging

from fastapi import APIRouter
from pydantic import BaseModel, Field

from ant31box.version import VERSION

router = APIRouter()

logger = logging.getLogger(__name__)


class VersionResp(BaseModel):
    version: str = Field(...)


@router.get("/", tags=["info"])
async def index():
    return await version()


@router.get("/version", tags=["info"], response_model=VersionResp)
async def version() -> VersionResp:
    return VersionResp(version=VERSION.app_version)



# pylint: disable=no-self-argument

import json
import logging
import logging.config
import os
from typing import Any, Generic, Self, TypeVar

import yaml
from pydantic import BaseModel, ConfigDict, Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from ant31box.utilsd import deepmerge

LOG_LEVELS: dict[str, int] = {
    "critical": logging.CRITICAL,
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
}

LOGGING_CONFIG: dict[str, Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "ant31box.logutils.DefaultFormatter",
            "fmt": "%(levelprefix)s %(message)s",
            "use_colors": None,
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
            "level": "INFO",
        },
    },
    "loggers": {
        "ant31box": {"handlers": ["default"], "level": "INFO", "propagate": True},
    },
}


logger: logging.Logger = logging.getLogger("ant31box")


class BaseConfig(BaseModel):
    model_config = ConfigDict(extra="allow")


class AppConfigSchema(BaseConfig):
    env: str = Field(default="dev")
    prometheus_dir: str = Field(default="/tmp/prometheus")


class CorsConfigSchema(BaseConfig):
    allow_origin_regex: str = Field(default=".*")
    allow_origins: list[str] = Field(
        default=[
            "http://localhost:8080",
            "http://localhost:8000",
            "http://localhost",
        ]
    )
    allow_credentials: bool = Field(default=False)
    allow_methods: list[str] = Field(default=["*"])
    allow_headers: list[str] = Field(default=["*"])


class TokenAuthMiddleWare(BaseConfig):
    token: str = Field(default="")
    parameter: str = Field(default="token_auth")
    header_name: str = Field(default="token")
    skip_paths: list[str] = Field(
        default=[
            "/api/v1/auth/login",
            "/api/v1/auth/register",
            "/docs",
            "/openapi.json",
            "/redoc",
            "/metrics",
            "/health",
        ]
    )


class FastAPIConfigSchema(BaseConfig):
    server: str = Field(default="ant31box.server.server:serve")
    middlewares: list[str] = Field(default_factory=list)
    routers: list[str] = Field(default_factory=list)
    middlewares_replace_default: list[str] | None = Field(
        default=None, description="Replace default middlewares, if set to None, default middlewares are added"
    )
    routers_replace_default: list[str] | None = Field(
        default=None, description="Replace default routers, if set to None, default routers are added"
    )
    cors: CorsConfigSchema = Field(default_factory=CorsConfigSchema)
    token_auth: TokenAuthMiddleWare = Field(default_factory=TokenAuthMiddleWare)
    token: str = Field(default="")
    host: str = Field(default="0.0.0.0")
    port: int = Field(default=8080)
    reload: bool = Field(default=False)

    @field_validator("port")
    def convert_port(cls, v) -> int:
        return int(v)


class LoggingConfigSchema(BaseConfig):
    use_colors: bool = Field(default=True)
    log_config: dict[str, Any] | str | None = Field(default_factory=lambda: LOGGING_CONFIG)
    level: str = Field(default="info")


class SentryConfigSchema(BaseConfig):
    dsn: str | None = Field(default=None)
    environment: str | None = Field(default=None)
    release: str | None = Field(default=None)
    traces_sample_rate: float | None = Field(default=None)


ENVPREFIX = "ANT31BOX"


class S3ConfigSchema(BaseConfig):
    endpoint: str = Field(default="https://s3.eu-central-1.amazonaws.com")
    access_key: str = Field(default="")
    secret_key: str = Field(default="")
    region: str = Field(default="eu-central-1")
    prefix: str = Field(default="")
    bucket: str = Field(default="abucket")


# Main configuration schema
class ConfigSchema(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix=f"{ENVPREFIX}_", env_nested_delimiter="__", case_sensitive=False, extra="allow"
    )
    app: AppConfigSchema = Field(default_factory=AppConfigSchema)
    logging: LoggingConfigSchema = Field(default_factory=LoggingConfigSchema)
    server: FastAPIConfigSchema = Field(default_factory=FastAPIConfigSchema)
    sentry: SentryConfigSchema = Field(default_factory=SentryConfigSchema)
    name: str = Field(default="ant31box")


TBaseConfig = TypeVar("TBaseConfig", bound=BaseSettings)  # pylint: disable= invalid-name


class GenericConfig(Generic[TBaseConfig]):
    _env_prefix = ENVPREFIX
    __config_class__: type[TBaseConfig]

    def __init__(self, conf: TBaseConfig):
        self.loaded = False
        self._conf: TBaseConfig = conf
        self._set_conf(conf)

    @property
    def conf(self) -> TBaseConfig:
        return self._conf

    def _set_conf(self, conf: TBaseConfig) -> None:
        self._conf = conf
        self.load(force=True)

    def load(self, force=True) -> bool:
        if not self.loaded or force:
            self.loaded = True
            return True
        raise RuntimeError("Config already loaded")

    @classmethod
    def from_yaml(cls, file_path: str) -> Self:
        with open(file_path, encoding="utf-8") as file:
            config_dict = yaml.safe_load(file)
        # merge init + env config
        alreadyinit = cls.__config_class__().model_dump(exclude_unset=True, exclude_defaults=True)
        deepmerge(config_dict, alreadyinit)
        logger.debug("Config loaded from %s: %s", file_path, config_dict)

        return cls(cls.__config_class__.model_validate(config_dict))

    @classmethod
    def default_config(cls) -> Self:
        return cls(cls.__config_class__())

    @classmethod
    def auto_config(cls, path: str | None = None) -> Self:
        if path:
            paths = [path]
        else:
            paths = [
                os.environ.get(f"{cls._env_prefix}_CONFIG", "localconfig.yaml"),
                "config.yaml",
            ]
        conf = cls.default_config()
        matched = False
        for p in paths:
            if os.path.exists(p):
                conf = cls.from_yaml(p)
                conf.load()
                logger.info("Config loaded: %s", p)
                matched = True
                break
        if not matched:
            logger.warning("No config file found, using default config")
            conf.load()
        return conf

    def replace(self, other: TBaseConfig) -> None:
        self._set_conf(other)

    def dump(self, destpath: str = "") -> str:
        dump = yaml.dump(self.conf.model_dump())
        if destpath:
            with open(destpath, "w", encoding="utf-8") as file:
                file.write(dump)
        return dump


TConfigSchema = TypeVar("TConfigSchema", bound=ConfigSchema)  # pylint: disable= invalid-name


class Config(Generic[TConfigSchema], GenericConfig[TConfigSchema]):
    __config_class__: type[TConfigSchema]

    @property
    def logging(self) -> LoggingConfigSchema:
        return self.conf.logging

    @property
    def server(self) -> FastAPIConfigSchema:
        return self.conf.server

    @property
    def sentry(self) -> SentryConfigSchema:
        return self.conf.sentry

    @property
    def app(self) -> AppConfigSchema:
        return self.conf.app

    @property
    def name(self) -> str:
        return self.conf.name

    def load(self, force=True) -> bool:
        if not self.loaded or force:
            self.configure_logging()
            self.loaded = True
            return True
        raise RuntimeError("Config already loaded")

    def configure_logging(self) -> None:
        log_config = self.logging.log_config
        use_colors = self.logging.use_colors
        log_level = self.logging.level

        if log_config:
            loaded_config = None
            if isinstance(log_config, dict):
                if use_colors in (True, False):
                    log_config["formatters"]["default"]["use_colors"] = use_colors
                loaded_config = log_config

            elif log_config.endswith(".json"):
                with open(log_config, encoding="utf-8") as file:
                    loaded_config = json.load(file)

            elif log_config.endswith((".yaml", ".yml")):
                with open(log_config, encoding="utf-8") as file:
                    loaded_config = yaml.safe_load(file)
            else:
                # See the note about fileConfig() here:
                # https://docs.python.org/3/library/logging.config.html#configuration-file-format
                logging.config.fileConfig(log_config, disable_existing_loggers=False)
            if loaded_config is not None:
                if (
                    "loggers" in loaded_config
                    and self.name not in loaded_config["loggers"]
                    and "ant31box" in loaded_config["loggers"]
                ):
                    loaded_config["loggers"][self.name] = loaded_config["loggers"]["ant31box"]
                logging.config.dictConfig(loaded_config)

        if log_level is not None:
            if isinstance(log_level, str):
                log_level = LOG_LEVELS[log_level.lower()]
            logging.getLogger(self.name).setLevel(log_level)
            logging.getLogger("ant31box").setLevel(log_level)
            logging.getLogger("root").setLevel(log_level)


T = TypeVar("T", bound=GenericConfig)


# Singleton to get the Configuration instance
# pylint: disable=super-init-not-called
class GConfig(Generic[T]):
    __instance__: T | None = None
    __conf_class__: type[T] | None = None

    def __init__(self, path: str | None = None):
        self.path = path

    def __new__(cls, path: str | None = None) -> T:
        if cls.__instance__ is None:
            if cls.__conf_class__ is None:
                raise ValueError("conf_class not set")
            cls.__instance__ = cls.__conf_class__.auto_config(path)
        return cls.__instance__

    @classmethod
    def reinit(cls) -> None:
        cls.__instance__ = None

    @classmethod
    def set_conf_class(cls, conf_class: type[T]) -> None:
        cls.__conf_class__ = conf_class

    @classmethod
    def instance(cls) -> T:
        if cls.__instance__ is None:
            raise ValueError("Instance not initialized")
        return cls.__instance__


def config(path: str | None = None, confclass: type[T] = Config, reload: bool = False) -> T:
    if confclass is not None:
        GConfig[confclass].set_conf_class(confclass)
    if reload:
        GConfig[confclass].reinit()
    # load the configuration
    _ = GConfig[confclass](path)
    return GConfig[confclass].instance()
